# This seems like a reasonable query for right now.
# TODO: make better (Google has been working on theirs since 1994)

def GOLDENQ(query, inner_hit_size=10):
    return {
        'query': {
            'bool': {
                'should': [
                    {
                        'nested': {
                            'path': 'captions',
                            'query': {
                                'match': {
                                    'captions.content': query
                                }
                            },
                            'inner_hits': {
                                'size': inner_hit_size,
                                'highlight': {
                                    'fields': {
                                        'captions.content': {}
                                    }
                                }
                            }
                        }
                    },
                    {
                        'match': {
                            'description': query
                        }
                    }
                ]
            }
        },
        'highlight': {
            'fields': {
                'description': {}
            }
        }
    }



from django.http import JsonResponse, Http404
from django.conf import settings

from .indexing import Index

def clean_hit( hit ):
    captions = [ dict( (*caption['_source'].items(), ('highlights', caption['highlight']['captions.content']), ('score', caption['_score'])) ) for caption in hit['inner_hits']['captions']['hits']['hits']]
    return dict( (*hit['_source'].items(), ('highlights', hit['highlight']['description'] if 'highlight' in hit.keys() else []), ('score', hit['_score']), ('captions',captions)) )

def search(request, index, query):
    if index not in settings.YTCI_SETTINGS.keys():
        raise Http404('Index %s does note exist'%index)

    p = int(request.GET.get('p', '0'))
    idx = Index(**settings.YTCI_SETTINGS[index]['elastic'])
    r = idx.search(query, from_=(p*20))
    results = [ clean_hit(hit) for hit in r['hits']['hits'] ]
    ctx = {'results':results, 'query': query, 'index':index}
    return JsonResponse(ctx)

def count(request, index):
    idx = Index(**settings.YTCI_SETTINGS[index]['elastic'])
    return JsonResponse({'index': index, 'count': idx.count()['count']})



from elasticsearch import Elasticsearch
from elasticsearch import helpers

class ESIndexError(Exception):
    pass

class ESIndex:
    # ESIndex is a generic Elasticsearch helper class
    def __init__(self, hosts=['localhost:9200'], create=True):
        self._es = Elasticsearch(hosts)
        if create and not self._exits():
            self._create_index()

    def _index_exists(func):
        def wrapper(self, *args, **kwargs):
            if not self._exits(): raise ESIndexError('Index %s does not exist'%self.Meta.index_name)
            return func(self, *args, **kwargs)
        return wrapper

    def _exits(self):
        return self._es.indices.exists(self.Meta.index_name)

    def _create_index(self):
        if self._exits(): raise ESIndexError('Index %s already exists'%self.Meta.index_name)
        self._es.indices.create(self.Meta.index_name)
        self._es.indices.put_mapping(index=self.Meta.index_name, doc_type=self.Meta.doc_type, body=self.Meta.mapping)

    @_index_exists
    def _delete_index(self, confirm=None):
        # This may seem stupid but be very sure before you delete the index.
        if confirm == 'DELETE':
            self._es.indices.delete(index=self.Meta.index_name)

    @_index_exists
    def index_object(self, obj):
        return self._es.index(index=self.Meta.index_name, id=self.Meta.doc_id_tmpl%obj, body=obj)

    @_index_exists
    def bulk(self, oblist):
        bulk_list = [
            {
                '_index': self.Meta.index_name,
                '_id': self.Meta.doc_id_tmpl%ob,
                '_type': self.Meta.doc_type,
                '_source': ob
            } for ob in oblist
        ]
        helpers.bulk(self._es, bulk_list)

    @_index_exists
    def count(self):
        return self._es.count(index=self.Meta.index_name)



from urllib import request
from xml.etree import ElementTree

from django.core.management.base import BaseCommand, CommandError
from django.conf import settings

from ytindex.indexing import Index
from ytindex.indexing.ytapi import Youtube
from ytindex.indexing.downloader import YTCaptionDownloader, YTCaptionNotFoundException

class Command(BaseCommand):
    help = 'Index the latest videos from a channel'

    def add_arguments(self, parser):
        parser.add_argument('index', type=str)
        parser.add_argument('playlist_ids', nargs='+', type=str)

    def handle(self, *args, **options):
        index = options['index']
        idx_settings = settings.YTCI_SETTINGS[index]

        def index_video(ytid, idx, ls=12):
            downloader = YTCaptionDownloader(ytid, line_size=ls)
            idx.index_object(downloader.as_dict())

        for playlist_id in options['playlist_ids']:
            idx = Index(**idx_settings['elastic'])
            y = Youtube( settings.YTCI_SETTINGS['api_key'] )
            for video_id in y.listplaylist(playlist_id):
                try:
                    print('Indexing::', video_id)
                    index_video(video_id, idx, ls=idx_settings.get('line_size', 12))
                except YTCaptionNotFoundException as e:
                    print(e)
                except Exception as e:
                    print('Most unexpected!!!', e)



import re
from datetime import datetime, timedelta

# It might make sense to use the official API down the line
# For now this was quicker than getting a key, etc.
from pytube import YouTube

class YTCaptionException(Exception):
    pass
class YTCaptionNotFoundException(YTCaptionException):
    pass

class YTCaptionDownloader:
    # CaptionDownloader downloads video info and auto captions for a given
    # Youtube video and makes it available as a dict.

    def __init__(self, yt_id, line_size=20, lang='en'):
        # yt_id: id of the Youtube video
        # line_size: how many individual caption lines should be chunked together
        # lang: desired subtitle language. (hardly anything but en makes sense right now but who knows)
         self.yt_id = yt_id
         self._ls = line_size
         self._lang = lang
         self._captions = None
         self.__ud = None
         self.__create_yt()

    def __create_yt(self):
        ## pytube.YouTube sometimes fails on __init__ idkw
        MAX_RETRIES = 3
        retries = 0
        while retries < MAX_RETRIES:
            try:
                self._yt = YouTube('https://www.youtube.com/watch?v=%s'%self.yt_id)
                return
            except Exception as e:
                retries += 1
                print('pytube.YouTube failed on init %i tries left'%(MAX_RETRIES-retries))
                if retries >= MAX_RETRIES:
                    print('Catastrophic: pytube.Youtube failed on init %i times. Giving up.'%MAX_RETRIES)
                    raise e

    @classmethod
    def timecode2seconds(cls, timecode):
        # SRT timecodes are HH:MM:SS, we want seconds
        tr = re.compile('(?P<hours>\d{2}):(?P<minutes>\d{2}):(?P<seconds>\d{2})')
        td = timedelta( **dict( ((k,int(v)) for k, v in tr.match(timecode).groupdict().items()) ) )
        return td.total_seconds()

    @classmethod
    def compact(cls, lines, line_size=5 ):
        # Captions can be super atomic, group them together.

        def _merge_lines( lines ):
            # Merge a list of lines into one
            start = lines[0]['start']
            end = lines[-1]['end']
            content = ' '.join([l['content'] for l in lines])
            return {'start': start, 'end': end, 'content': content}

        i = 0
        new_lines = []
        while i < len(lines)-line_size:
            new_lines.append(_merge_lines(lines[i:i+line_size]))
            i += line_size

        # Merge any remaining lines
        new_lines.append(_merge_lines(lines[i:]))
        return new_lines

    def _get_captions(self, lang='en', line_size=20 ):
        c = self._yt.captions.get_by_language_code(lang)
        if not c: raise YTCaptionNotFoundException('No captions found for %s'%self.id)
        s = c.generate_srt_captions()
        s = re.sub( '<[^>]*>', '', s ) # Remove styling
        lines = s.split('\n\n') # Split the SRT sections

        # Parse SRT
        def _parse_srt_line(line):
            index, time, content = line.split('\n')
            start, end = time.split(' --> ')
            start = self.timecode2seconds(start)
            end = self.timecode2seconds(end)
            return {'start':start, 'end':end, 'content':content}

        lines = [_parse_srt_line(line) for line in lines]
        return self.compact(lines, line_size=line_size)

    # Properties
    @property
    def id(self):
        return self._yt.video_id
    @property
    def author(self):
        return self._yt.player_config_args['author']
    @property
    def title(self):
        return self._yt.title
    @property
    def description(self):
        return self._yt.description
    @property
    def thumbnail(self):
        return self._yt.thumbnail_url
    @property
    def upload_date(self):
        if self.__ud is None:
            reg = r'(?isx)<meta(?=[^>]+(?:itemprop|name|property|id|http-equiv)=(["\']?)datePublished\1)[^>]+?content=(["\'])(?P<upload_date>.*?)\2'
            m = re.search(reg, self._yt.watch_html)
            if m:
                date = m.group('upload_date')
                self.__ud = datetime.strptime(date, '%Y-%m-%d')
        return self.__ud
    @property
    def captions(self):
        if self._captions is None:
            self._captions = self._get_captions( lang=self._lang, line_size=self._ls)
        return self._captions
    @property
    def url(self):
        return self._yt.watch_url
    def jump_url(self, seconds):
        return '%s&t=%ss'%(self.url, seconds)

    def as_dict(self):
        return {
            'id': self.id,
            'title': self.title,
            'description': self.description,
            'thumbnail': self.thumbnail,
            'url': self.url,
            'upload_date': self.upload_date,
            'captions': self.captions
        }



from .ytcaptions import YTCaptionIndex

def index_class_factory( es_index_class=YTCaptionIndex, index='yt-captions', dtype='yt-video', doc_id_tmpl='yt-video-%(id)s' ):
    # We may want to index different Youtube channels using differnt
    # index and doc type names?
    class index_class(es_index_class):
        pass
    index_class.Meta.index_name = index
    index_class.Meta.doc_type = dtype
    index_class.Meta.doc_id_tmpl = doc_id_tmpl
    return index_class

def Index( index, dtype='yt-video', es_host='localhost:9200' ):
    # Create and instantiate a dynamic YTCaptionIndex
    return index_class_factory(index=index, dtype=dtype)(es_host)



from urllib import request
from xml.etree import ElementTree

from django.core.management.base import BaseCommand, CommandError
from django.conf import settings

from ytindex.indexing import Index
from ytindex.indexing.downloader import YTCaptionDownloader, YTCaptionNotFoundException

class Command(BaseCommand):
    help = 'Index the latest videos from a channel'

    def add_arguments(self, parser):
        parser.add_argument('index', nargs='+', type=str)

    def handle(self, *args, **options):

        def index_video(ytid, idx, ls=12):
            downloader = YTCaptionDownloader(ytid, line_size=ls)
            idx.index_object(downloader.as_dict())

        for index in options['index']:
            idx_settings = settings.YTCI_SETTINGS[index]
            idx = Index(**idx_settings['elastic'])
            channel_id = idx_settings['channel_id']

            r = request.urlopen('https://www.youtube.com/feeds/videos.xml?channel_id=%s'%channel_id)
            if r.status == 200:
                et = ElementTree.fromstring(r.read())
                ns = {'xmlns':'http://www.w3.org/2005/Atom'}
                xpath = './xmlns:entry/xmlns:id'
                video_ids = [id.text.split(':')[-1] for id in et.findall( xpath, namespaces=ns )]
                for video_id in video_ids:
                    try:
                        print('Indexing::', video_id)
                        index_video(video_id, idx, ls=idx_settings.get('line_size', 12))
                    except YTCaptionNotFoundException as e:
                        print(e)
                    except Exception as e:
                        print('Most unexpected!!!', e)



from django.urls import path, re_path
from .views import search, count

urlpatterns = [
    re_path('(?P<index>\w*)/_count', count, name='ytindex-count'),
    re_path('(?P<index>\w*)/(?P<query>.*)', search, name='ytindex-search'),
]



from .esindex import ESIndex
from .golden_q import GOLDENQ

class YTCaptionIndex(ESIndex):
    # YTCaptionIndex defines an ESIndex specific
    # to indexing and searching Youtube video captions
    class Meta:
        index_name = 'yt-captions'
        doc_type = 'yt-video'
        doc_id_tmpl = 'yt-video-%(id)s'
        mapping = {
          'properties': {
            'id': { 'type': 'text', 'index': 'false' },
            'upload_date': { 'type': 'date', 'index': 'false'  },
            'title': { 'type': 'text' },
            'description': { 'type': 'text', 'analyzer':'english' },
            'thumbnail': { 'type': 'text', 'index': 'false'  },
            'url': { 'type': 'text', 'index': 'false' },
            'captions': {
              'type': 'nested',
              'properties': {
                'start': {'type':'long', 'index': 'false' },
                'end': {'type':'long', 'index': 'false' },
                'content': {'type':'text', 'analyzer':'english'}
              }
            }
          }
        }

    def search(self, query, query_type='match_phrase', size=20, from_=0, q_func=GOLDENQ):
        q = q_func(query)
        return self._es.search(index=self.Meta.index_name, body=q, size=size, from_=from_)

    def get(self, id):
        return self._es.get(id=self.Meta.doc_id_tmpl%{'id':id}, index=self.Meta.index_name)



import os
import googleapiclient.discovery
import googleapiclient.errors

os.environ["OAUTHLIB_INSECURE_TRANSPORT"] = "1"
scopes = ["https://www.googleapis.com/auth/youtube.readonly"]

class Youtube():
  def __init__(self, key):
    api_service_name = "youtube"
    api_version = "v3"
    self.__yt = googleapiclient.discovery.build(api_service_name, api_version, developerKey=key)

  def listplaylist(self, id):
    r = self.__yt.playlistItems().list(part='contentDetails',playlistId=id, maxResults=50).execute()
    for result in r.get('items', []):
      yield result['contentDetails']['videoId']

    while 'nextPageToken' in r.keys():
      r = self.__yt.playlistItems().list(part='contentDetails',playlistId=id, maxResults=50, pageToken=r['nextPageToken']).execute()
      for result in r.get('items', []):
        yield result['contentDetails']['videoId']
